import yaml
import marshmallow_dataclass
import json

from typing import Dict, Any, Optional
from logging import getLogger

from dataclasses import dataclass, field
from marshmallow.exceptions import MarshmallowError, ValidationError
from abc import ABC

ENCAB = "encab"
ENCAB_GELF = "encab_gelf"

mylogger = getLogger(ENCAB_GELF)


class ConfigError(ValueError):
    pass


@dataclass
class RecognizerSettings(ABC):
    type: str = field(default="default")
    # recognizer ``default`` or ``grok`` or ``json``
    pattern: Optional[str] = field(default=None)
    # grok pattern if ``type == 'grok'``
    #
    # see: https://pypi.org/project/pygrok/


@dataclass
class GelfHandlerSettings(ABC):
    protocol: str
    # GELF Protocol. One of HTTP, HTTPS, UDP, TCP, TLS
    host: str
    # Graylog input host or IP address (default=localhost)
    port: int = field(default=12201)
    # Graylog port (default=12201)
    optional_fields: Dict[str, Any] = field(default_factory=lambda: dict())
    # optional gelf fields as dictionary. Will be added to each log record.

    # -- HTTP

    path: str = field(default="/gelf")
    # ('/gelf' by default) - path of the HTTP input
    compress: bool = field(default=True)
    # (True by default) - if true, compress log messages before sending them to the server
    timeout: float = field(default=5.0)
    # (5 by default) - amount of seconds that HTTP client should wait before it discards the request if the server doesn't respond

    # -- UDP

    # (True by default) - if true, compress log messages before sending them to the server
    chunk_size: int = field(default=1300)
    # (1300 by default) - maximum length of the message. If log length exceeds this value, it splits into multiple chunks

    # -- HTTPS

    validate: bool = field(default=False)
    # if true, validate server certificate. In that case ca_certs are required
    ca_certs: Optional[str] = field(default=None)
    # path to CA bundle file. For instance, on CentOS it would be '/etc/pki/tls/certs/ca-bundle.crt'

    # -- TLS

    # path to CA bundle file. For instance, on CentOS it would be '/etc/pki/tls/certs/ca-bundle.crt'
    certfile: Optional[str] = field(default=None)
    # path to the certificate file that is used to identify ourselves to the server
    keyfile: Optional[str] = field(default=None)
    # path to the private key. If the private key is stored with the certificate, this parameter can be ignored

    recognizer: RecognizerSettings = field(default_factory=lambda: RecognizerSettings())
    # log line recognizer settings


@dataclass
class GelfSettings(object):
    """
    the Graylog/GELF settings

    example:

    .. code-block:: yaml

        encab:
            debug: true
            halt_on_exit: False
        extensions:
            gelf:
                module: encab_gelf
                enabled: true
                settings:
                    handlers:
                        HTTP:
                            protocol: HTTP
                            host: localhost # host name or IP address of the GELF input
                            port: 80        # port of the GELF input
                            optional_fields: # optiona fields added to each log record 
                                localname: encab
    """

    handlers: Dict[str, GelfHandlerSettings]

    @staticmethod
    def update_default_handler(
        settings: "GelfSettings", environment: Dict[str, Any]
    ) -> "GelfSettings":

        GRAYLOG_PROTOCOL = "GRAYLOG_PROTOCOL"
        GRAYLOG_HOST = "GRAYLOG_HOST"
        GRAYLOG_PORT = "GRAYLOG_PORT"
        GRAYLOG_OPTIONAL_FIELDS = "GRAYLOG_OPTIONAL_FIELDS"

        set_default_handler = False
        for var in [
            GRAYLOG_PROTOCOL,
            GRAYLOG_HOST,
            GRAYLOG_PORT,
            GRAYLOG_OPTIONAL_FIELDS,
        ]:
            if var in environment:
                set_default_handler = True
                break

        if set_default_handler:
            optional_fields_var = environment.get(GRAYLOG_OPTIONAL_FIELDS)
            optional_fields = dict()
            if optional_fields_var:
                try:
                    optional_fields = json.loads(optional_fields_var)
                    assert isinstance(optional_fields, dict)
                except Exception as e:
                    raise ConfigError(
                        f"Error parsing environment variable {GRAYLOG_OPTIONAL_FIELDS}: {str(e)}"
                    )

            if "default" not in settings.handlers:
                protocol = environment.get(GRAYLOG_PROTOCOL, "HTTP")
                host = environment.get(GRAYLOG_HOST, "localhost")
                port = environment.get(GRAYLOG_PORT, 12201)
                settings.handlers["default"] = GelfHandlerSettings(
                    protocol, host, port, optional_fields
                )
            else:
                handler = settings.handlers["default"]
                handler.protocol = environment.get(GRAYLOG_PROTOCOL, handler.protocol)
                handler.host = environment.get(GRAYLOG_HOST, handler.host)
                handler.port = environment.get(GRAYLOG_PORT, handler.port)
                handler.optional_fields = (
                    optional_fields if optional_fields_var else handler.optional_fields
                )

        return settings

    @staticmethod
    def load(settings: Dict[str, Any]) -> "GelfSettings":
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(GelfSettings)
            settings = ConfigSchema().load(settings)  # type: ignore
            assert isinstance(settings, GelfSettings)
            return settings
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{ENCAB_GELF}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)



#
# Derived from https://github.com/keeprocking/pygelf/blob/master/pygelf/gelf.py
#

import datetime
import logging
import json
import zlib
import os
import struct
import traceback


LEVELS = {
    logging.DEBUG: 7,
    logging.INFO: 6,
    logging.WARNING: 4,
    logging.ERROR: 3,
    logging.CRITICAL: 2,
}


# skip_list is used to filter additional fields in a log message.
# It contains all attributes listed in
# http://docs.python.org/library/logging.html#logrecord-attributes
# plus exc_text, which is only found in the logging module source,
# and id, which is prohibited by the GELF format.
SKIP_LIST = (
    "args",
    "asctime",
    "created",
    "exc_info",
    "exc_text",
    "filename",
    "funcName",
    "id",
    "levelname",
    "levelno",
    "lineno",
    "module",
    "msecs",
    "message",
    "msg",
    "name",
    "pathname",
    "process",
    "processName",
    "relativeCreated",
    "thread",
    "threadName",
)


def make(
    record,
    domain,
    debug,
    version,
    additional_fields,
    additional_env_fields,
    include_extra_fields=False,
):
    gelf = {
        "version": version,
        "short_message": record.getMessage(),
        "timestamp": record.created,
        "level": LEVELS[record.levelno],
        "host": domain,
    }

    if record.exc_info:
        gelf["full_message"] = "\n".join(traceback.format_exception(*record.exc_info))
    elif record.exc_text is not None:
        # QueueHandler, if used, formats the record, so that exc_info will always be empty:
        # https://docs.python.org/3/library/logging.handlers.html#logging.handlers.QueueHandler
        gelf["full_message"] = record.exc_text

    if debug:
        gelf["_file"] = record.filename
        gelf["_line"] = record.lineno
        gelf["_module"] = record.module
        gelf["_func"] = record.funcName
        gelf["_logger_name"] = record.name

    if additional_fields is not None:
        gelf.update(additional_fields)

    if additional_env_fields is not None:
        appended = {}
        for name, env in additional_env_fields.items():
            if env in os.environ:
                appended["_" + name] = os.environ.get(env)

        gelf.update(appended)

    if include_extra_fields:
        add_extra_fields(gelf, record)

    return gelf


def add_extra_fields(gelf, record):
    for key, value in record.__dict__.items():
        if key not in SKIP_LIST and not key.startswith("_"):
            gelf["_%s" % key] = value


def object_to_json(obj):
    """Convert object that cannot be natively serialized by python to JSON representation."""
    if isinstance(obj, (datetime.datetime, datetime.date, datetime.time)):
        return obj.isoformat()
    return str(obj)


def pack(gelf, compress, default):
    packed = json.dumps(gelf, separators=(",", ":"), default=default).encode("utf-8")
    return zlib.compress(packed) if compress else packed


def split(gelf, chunk_size):
    header = b"\x1e\x0f"
    message_id = os.urandom(8)
    chunks = [gelf[pos : pos + chunk_size] for pos in range(0, len(gelf), chunk_size)]
    number_of_chunks = len(chunks)

    for chunk_index, chunk in enumerate(chunks):
        yield b"".join(
            (
                header,
                message_id,
                struct.pack("b", chunk_index),
                struct.pack("b", number_of_chunks),
                chunk,
            )
        )



#
# Derived from https://github.com/keeprocking/pygelf/blob/master/pygelf/handlers.py
#

import ssl
import socket

import http.client as httplib
from logging.handlers import SocketHandler, DatagramHandler
from logging import Handler as LoggingHandler
from . import gelf


class BaseHandler(object):
    def __init__(
        self,
        debug=False,
        version="1.1",
        include_extra_fields=False,
        compress=False,
        static_fields=None,
        json_default=gelf.object_to_json,
        additional_env_fields=None,
        **kwargs
    ):
        """
        Logging handler that transforms each record into GELF (graylog extended log format) and sends it over TCP.

        :param debug: include debug fields, e.g. line number, or not
        :param include_extra_fields: include non-default fields from record to message, or not
        :param json_default: function that is called for objects that cannot be serialized to JSON natively by python
        :param kwargs: additional fields that will be included in the log message, e.g. application name.
                       Each additional field should start with underscore, e.g. _app_name
        """

        self.additional_env_fields = additional_env_fields
        if self.additional_env_fields is None:
            self.additional_env_fields = {}

        self.debug = debug
        self.version = version
        self.additional_fields = static_fields if static_fields else kwargs
        self.include_extra_fields = include_extra_fields
        self.additional_fields.pop("_id", None)
        self.domain = socket.gethostname()
        self.compress = compress
        self.json_default = json_default

    def convert_record_to_gelf(self, record):
        additional_fields = self.additional_fields or dict()
        if hasattr(record, "extra") and isinstance(record.extra, dict):
            additional_fields = {**additional_fields, **record.extra}

        return gelf.pack(
            gelf.make(
                record,
                self.domain,
                self.debug,
                self.version,
                additional_fields,
                self.additional_env_fields,
                self.include_extra_fields,
            ),
            self.compress,
            self.json_default,
        )


class GelfTcpHandler(BaseHandler, SocketHandler):

    def __init__(self, host, port, **kwargs):
        """
        Logging handler that transforms each record into GELF (graylog extended log format) and sends it over TCP.

        :param host: GELF TCP input host
        :param port: GELF TCP input port
        """

        SocketHandler.__init__(self, host, port)
        BaseHandler.__init__(self, **kwargs)

    def makePickle(self, record):
        """if you send the message over tcp, it should always be null terminated or the input will reject it"""
        return self.convert_record_to_gelf(record) + b"\x00"


class GelfUdpHandler(BaseHandler, DatagramHandler):

    def __init__(self, host, port, compress=True, chunk_size=1300, **kwargs):
        """
        Logging handler that transforms each record into GELF (graylog extended log format) and sends it over UDP.
        If message length exceeds chunk_size, the message splits into multiple chunks.
        The number of chunks must be less than 128.

        :param host: GELF UDP input host
        :param port: GELF UDP input port
        :param compress: compress message before sending it to the server or not
        :param chunk_size: length of a chunk, should be less than the MTU (maximum transmission unit)
        """

        DatagramHandler.__init__(self, host, port)
        BaseHandler.__init__(self, compress=compress, **kwargs)

        self.chunk_size = chunk_size

    def send(self, s):
        if len(s) <= self.chunk_size:
            DatagramHandler.send(self, s)
            return

        chunks = gelf.split(s, self.chunk_size)
        for chunk in chunks:
            DatagramHandler.send(self, chunk)

    def makePickle(self, record):
        return self.convert_record_to_gelf(record)


class GelfTlsHandler(GelfTcpHandler):

    def __init__(
        self,
        host,
        port,
        validate=False,
        ca_certs=None,
        certfile=None,
        keyfile=None,
        **kwargs
    ):
        """
        TCP GELF logging handler with TLS support

        :param host: GELF UDP input host
        :param port: GELF UDP input port
        :param validate: if true, validate server certificate. In that case ca_certs are required
        :param ca_certs: path to CA bundle file. For instance, on CentOS it would be '/etc/pki/tls/certs/ca-bundle.crt'
        :param certfile: path to the certificate file that is used to identify ourselves to the server
        :param keyfile: path to the private key. If the private key is stored with the certificate,
                        this parameter can be ignored
        """

        if validate and ca_certs is None:
            raise ValueError("CA bundle file path must be specified")

        if keyfile is not None and certfile is None:
            raise ValueError("certfile must be specified")

        GelfTcpHandler.__init__(self, host, port, **kwargs)

        self.ca_certs = ca_certs
        self.reqs = ssl.CERT_REQUIRED if validate else ssl.CERT_NONE
        self.certfile = certfile
        self.keyfile = keyfile if keyfile else certfile

    def makeSocket(self, timeout=1):
        plain_socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

        if hasattr(plain_socket, "settimeout"):
            plain_socket.settimeout(timeout)

        wrapped_socket = ssl.wrap_socket(
            plain_socket,
            ca_certs=self.ca_certs,
            cert_reqs=self.reqs,
            keyfile=self.keyfile,
            certfile=self.certfile,
        )
        wrapped_socket.connect((self.host, self.port))

        return wrapped_socket


class GelfHttpHandler(BaseHandler, LoggingHandler):

    def __init__(self, host, port, compress=True, path="/gelf", timeout=5, **kwargs):
        """
        Logging handler that transforms each record into GELF (graylog extended log format) and sends it over HTTP.

        :param host: GELF HTTP input host
        :param port: GELF HTTP input port
        :param compress: compress message before sending it to the server or not
        :param path: path of the HTTP input (http://docs.graylog.org/en/latest/pages/sending_data.html#gelf-via-http)
        :param timeout: amount of seconds that HTTP client should wait before it discards the request
                        if the server doesn't respond
        """

        LoggingHandler.__init__(self)
        BaseHandler.__init__(self, compress=compress, **kwargs)

        self.host = host
        self.port = port
        self.path = path
        self.timeout = timeout
        self.headers = {}

        if compress:
            self.headers["Content-Encoding"] = "gzip,deflate"

    def emit(self, record):
        data = self.convert_record_to_gelf(record)
        connection = httplib.HTTPConnection(
            host=self.host, port=self.port, timeout=self.timeout
        )
        try:
            connection.request("POST", self.path, data, self.headers)
        finally:
            connection.close()


class GelfHttpsHandler(BaseHandler, LoggingHandler):

    def __init__(
        self,
        host,
        port,
        compress=True,
        path="/gelf",
        timeout=5,
        validate=False,
        ca_certs=None,
        certfile=None,
        keyfile=None,
        keyfile_password=None,
        **kwargs
    ):
        """
        Logging handler that transforms each record into GELF (graylog extended log format) and sends it over HTTP.

        :param host: GELF HTTP input host
        :param port: GELF HTTP input port
        :param compress: compress message before sending it to the server or not
        :param path: path of the HTTP input (http://docs.graylog.org/en/latest/pages/sending_data.html#gelf-via-http)
        :param timeout: amount of seconds that HTTP client should wait before it discards the request
                        if the server doesn't respond
        :param validate: whether or not to validate the input's certificate
        :param ca_certs: path to the CA certificate file that signed the certificate the input is using
        :param certfile: not yet used
        :param keyfile: not yet used
        :param keyfile_password: not yet used
        """

        LoggingHandler.__init__(self)
        BaseHandler.__init__(self, compress=compress, **kwargs)

        self.host = host
        self.port = port
        self.path = path
        self.timeout = timeout
        self.headers = {}
        self.ca_certs = ca_certs
        self.keyfile = keyfile
        self.certfile = certfile
        self.keyfile_password = keyfile_password

        # Set up context: https://docs.python.org/3/library/http.client.html#http.client.HTTPSConnection
        # create_default_context returns an SSLContext object
        self.ctx = ssl.create_default_context()

        if validate and ca_certs is None:
            raise ValueError("CA bundle file path must be specified")

        if not validate:
            self.ctx.check_hostname = False
            self.ctx.verify_mode = ssl.CERT_NONE
        else:
            # Load our CA file
            self.ctx.load_verify_locations(cafile=self.ca_certs)

        if compress:
            self.headers["Content-Encoding"] = "gzip,deflate"

    def emit(self, record):
        data = self.convert_record_to_gelf(record)
        connection = httplib.HTTPSConnection(
            host=self.host, port=self.port, context=self.ctx, timeout=self.timeout
        )
        try:
            connection.request("POST", self.path, data, self.headers)
        finally:
            connection.close()



from typing import Dict, Optional
from abc import ABC, abstractmethod
from enum import Enum

from logging import getLogger, DEBUG

import re
from pygrok import Grok as GrokPattern  # type: ignore

mylogger = getLogger(__name__)
mylogger.setLevel(DEBUG)


class LogLevel(Enum):
    CRITICAL = 50
    EMERGENCY = CRITICAL
    ALERT = CRITICAL
    FATAL = CRITICAL
    ERROR = 40
    WARNING = 30
    WARN = WARNING
    INFO = 20
    DEBUG = 10

    @classmethod
    def levelsByName(cls) -> Dict[str, "LogLevel"]:
        return {
            "EMERGENCY": cls.EMERGENCY,
            "ALERT": cls.ALERT,
            "CRITICAL": cls.CRITICAL,
            "FATAL": cls.FATAL,
            "ERROR": cls.ERROR,
            "WARN": cls.WARNING,
            "WARNING": cls.WARNING,
            "INFO": cls.INFO,
            "DEBUG": cls.DEBUG,
        }

    @classmethod
    def namesBylevel(cls) -> Dict["LogLevel", str]:
        return {
            cls.CRITICAL: "CRITICAL",
            cls.ERROR: "ERROR",
            cls.WARNING: "WARNING",
            cls.INFO: "INFO",
            cls.DEBUG: "DEBUG",
        }

    @classmethod
    def fromString(cls, s: str) -> Optional["LogLevel"]:
        level = cls.levelsByName().get(s.upper(), 0)
        return LogLevel(level) if level else None

    @classmethod
    def levelPattern(cls) -> re.Pattern:
        names = cls.levelsByName().keys()
        pattern = "|".join([name for name in names])
        return re.compile(f"({pattern})", re.IGNORECASE)


class LogLine(object):
    def __init__(
        self,
        line: str,
        level: Optional[LogLevel] = None,
        attrs: Optional[Dict[str, str]] = None,
    ) -> None:
        self.level = level
        self.line = line
        self.attrs = attrs or dict()


class LogLineRecognizer(ABC):
    @abstractmethod
    def recognize(self, line: str) -> LogLine:
        pass


class GrokRecognizer(LogLineRecognizer):
    def __init__(self, pattern: str, log_level_tag: str = "LOGLEVEL") -> None:
        self.pattern = GrokPattern(pattern)
        self.log_level_tag = log_level_tag

    def recognize(self, line: str) -> LogLine:
        match = self.pattern.match(line)
        attrs = match.groupdict() if match else dict()
        attrs = {k: v for k, v in attrs.items() if k != self.log_level_tag}
        levelName = attrs.get(self.log_level_tag, 0)
        level = LogLevel.fromString(levelName) if levelName else None
        return LogLine(line, level, attrs)


class NoMatch(Exception):
    pass


class DefaultRecognizer(LogLineRecognizer):
    def __init__(self) -> None:
        self.levelPattern = LogLevel.levelPattern()

    def isNoLogLine(self, s: str) -> bool:
        return len(s) <= 1 and not (s[0].isalnum() or s[0] in ["[", "{", "|"])

    def recognize(self, line: str) -> LogLine:
        try:
            if self.isNoLogLine(line):
                raise NoMatch()

            match: Optional[re.Match[str]] = self.levelPattern.match(line)
            if not match:
                raise NoMatch()

            levelName = match.group(1)
            level = LogLevel.fromString(levelName) if levelName else None
            return LogLine(line, level)
        except NoMatch:
            return LogLine(line)



from typing import Dict, Any, Optional, Iterator

from logging import getLogger, Logger, Handler, DEBUG
from pluggy import HookimplMarker  # type: ignore

from .log_line_recognizer import LogLineRecognizer, DefaultRecognizer, GrokRecognizer
from .handlers import (
    MultiLineHandler,
    ErrorHandler,
    RecognizingHandler,
    ENCAB,
    ENCAB_GELF,
)

from .gelf.handlers import (
    GelfHttpHandler,
    GelfUdpHandler,
    GelfHttpsHandler,
    GelfTcpHandler,
    GelfTlsHandler,
)

from .config import RecognizerSettings, GelfHandlerSettings, GelfSettings, ConfigError

mylogger = getLogger(ENCAB_GELF)
mylogger.setLevel(DEBUG)


class RecognizerFactory(object):
    def __init__(self, settings: RecognizerSettings) -> None:
        self.settings = settings

    def create(self) -> LogLineRecognizer:
        if self.settings.type == "default":
            return DefaultRecognizer()
        elif self.settings.type == "grok":
            if not self.settings.pattern:
                raise ConfigError("Missing Grok regcognizer pattern")
            return GrokRecognizer(self.settings.pattern)
        else:
            raise ConfigError(f"Unsupported recognizer {self.settings.type}")


class GelfLogHandlerFactory(object):
    def __init__(self, gelf_settings: GelfSettings) -> None:
        self.gelf_settings = gelf_settings

    def update_settings(self, gelf_settings: GelfSettings) -> None:
        self.gelf_settings = gelf_settings

    def create(self, name: str, settings: GelfHandlerSettings):
        mylogger.info(
            f"Configuring GELF filter {name}: {str(settings)}",
            extra={"program": ENCAB_GELF},
        )
        assert settings.protocol in ["HTTP", "HTTPS", "UDP", "TCP", "TLS"]
        if settings.protocol == "HTTP":
            return GelfHttpHandler(
                host=settings.host,
                port=settings.port,
                compress=settings.compress,
                path=settings.path,
                timeout=settings.timeout,
                **settings.optional_fields,
            )
        elif settings.protocol == "HTTPS":
            return GelfHttpsHandler(
                host=settings.host,
                port=settings.port,
                compress=settings.compress,
                path=settings.path,
                timeout=settings.timeout,
                validate=settings.validate,
                ca_certs=settings.ca_certs,
                **settings.optional_fields,
            )
        elif settings.protocol == "UDP":
            return GelfUdpHandler(
                host=settings.host,
                port=settings.port,
                compress=settings.compress,
                chunk_size=settings.chunk_size,
                **settings.optional_fields,
            )
        elif settings.protocol == "TCP":
            return GelfTcpHandler(
                host=settings.host, port=settings.port, **settings.optional_fields
            )
        elif settings.protocol == "TLS":
            return GelfTlsHandler(
                host=settings.host,
                port=settings.port,
                validate=settings.validate,
                ca_certs=settings.ca_certs,
                certfile=settings.certfile,
                keyfile=settings.keyfile,
                **settings.optional_fields,
            )

    def createAll(self) -> Iterator[Handler]:
        for name, settings in self.gelf_settings.handlers.items():
            recognizers = RecognizerFactory(settings.recognizer)
            yield MultiLineHandler(
                name,
                RecognizingHandler(
                    ErrorHandler(self.create(name, settings), name),
                    name,
                    recognizers.create(),
                ),
            )


extension_impl = HookimplMarker(ENCAB)

is_enabled: bool = False

gelf_settings: Optional[GelfSettings] = None

factory: Optional[GelfLogHandlerFactory] = None


@extension_impl
def validate_extension(name: str, enabled: bool, settings: Dict[str, Any]):
    if name != ENCAB_GELF:
        return

    if not enabled:
        return

    gelf_settings = GelfSettings.load(settings)
    GelfLogHandlerFactory(gelf_settings)


@extension_impl
def configure_extension(name: str, enabled: bool, settings: Dict[str, Any]):
    global gelf_settings, factory, is_enabled

    if name != ENCAB_GELF:
        return

    is_enabled = enabled
    if not enabled:
        return

    gelf_settings = GelfSettings.load(settings)
    factory = GelfLogHandlerFactory(gelf_settings)


@extension_impl
def extend_environment(program_name: str, environment: Dict[str, str]):
    global gelf_settings, factory, is_enable

    if program_name == ENCAB:
        assert gelf_settings
        assert factory

        if "GRAYLOG_ENABLED" in environment:
            is_enabled = environment["GRAYLOG_ENABLED"] in ("True", "true", 1)

        if not is_enabled:
            return

        gelf_settings = GelfSettings.update_default_handler(gelf_settings, environment)
        factory.update_settings(gelf_settings)


@extension_impl
def update_logger(program_name: str, logger: Logger):
    global is_enable

    if program_name == ENCAB:
        return

    if not is_enabled:
        return

    if not factory:
        return

    mylogger.info("Adding GELF Handlers", extra={"program": ENCAB_GELF})

    for hander in factory.createAll():
        logger.addHandler(hander)



from typing import Callable
from threading import Thread, Lock, Event


class TimerClosed(Exception):
    pass


class OneShotTimer(object):

    STARTING = 0
    IDLE = 1
    RUNNING = 2
    SKIP = 3
    SKIPPING = 4
    CLOSING = 5
    CLOSED = 6

    def _waitStarted(self):
        self._state_change.wait()
        with self._lock:
            self._state_change.clear()
            if self._state == self.STARTING:
                self._state = self.RUNNING
            else:
                raise TimerClosed()

    def _runTarget(self):
        self._state_change.wait(self._timeout)
        skip = False
        with self._lock:
            self._state_change.clear()
            if self._state == self.RUNNING:
                pass
            elif self._state == self.SKIP:
                skip = True
                self._state = self.SKIPPING
            else:
                raise TimerClosed()
        if not skip:
            self._target()

    def _setIdle(self):
        with self._lock:
            if self._state in (self.RUNNING, self.SKIPPING):
                self._state = self.IDLE
            else:
                raise TimerClosed()

    def _run(self):
        try:
            while True:
                self._waitStarted()
                self._runTarget()
                self._setIdle()
        except TimerClosed:
            self._state = self.CLOSED

    def __init__(self, timeout: float, target: Callable) -> None:
        self._timeout = timeout
        self._target = target
        self._state_change = Event()
        self._state = self.IDLE
        self._lock = Lock()

        def timer():
            self._run()

        self._thread = Thread(target=timer, name="timer", daemon=True)
        self._thread.start()

    def start(self):
        with self._lock:
            if self._state == self.IDLE:
                self._state = self.STARTING
                self._state_change.set()

    def clear(self):
        with self._lock:
            if self._state == self.IDLE:
                self._state = self.SKIP
                self._state_change.set()

    def close(self):
        with self._lock:
            if self._state not in (self.CLOSING, self.CLOSED):
                self._state == self.CLOSING
                self._state_change.set()



from typing import Dict, Any, List, Optional

from logging import LogRecord, getLogger, Handler, DEBUG
from http.client import HTTPException

from .one_shot_timer import OneShotTimer
from .log_line_recognizer import LogLineRecognizer, DefaultRecognizer

ENCAB = "encab"
ENCAB_GELF = "encab_gelf"

mylogger = getLogger(__name__)
mylogger.setLevel(DEBUG)


class ExtLogRecord(LogRecord):
    def __init__(self, record: LogRecord) -> None:
        super().__init__(
            record.name,
            record.levelno,
            record.pathname,
            record.lineno,
            record.msg,
            record.args,
            record.exc_info,
            record.funcName,
            record.stack_info,
        )
        self.is_log_record = True
        self.extra: Dict[str, Any] = (
            record.extra
            if hasattr(record, "extra") and isinstance(record.extra, dict)
            else {}
        )

        self.suppress = self.extra.get("suppress", False)
        self.program = self.extra.get("program")
        self.is_from_encab = self.program in (ENCAB, ENCAB_GELF)

    @staticmethod
    def fromRecord(record: LogRecord) -> "ExtLogRecord":
        return record if isinstance(record, ExtLogRecord) else ExtLogRecord(record)

    def getMessage(self):
        msg = str(self.msg)
        try:
            if self.args:
                msg = msg % self.args
        except TypeError:
            pass
        return msg


class MultiLineHandler(Handler):
    TIMEOUT: float = 0.5

    def __init__(
        self, name: str, handler: Handler, timeout: Optional[float] = None
    ) -> None:
        super().__init__(handler.level)
        self.name = name
        self.handler = handler
        self.backlog: List[LogRecord] = list()
        self.timer = OneShotTimer(timeout or self.TIMEOUT, self.flush)

    def emit_upstream(self, record: LogRecord):
        self.handler.emit(record)

    def emitAll(self, record: LogRecord):
        self.timer.clear()
        self.flush()
        self.emit_upstream(record)

    def emit(self, log_record: LogRecord) -> None:
        record = ExtLogRecord.fromRecord(log_record)

        if not self.backlog:
            self.backlog.append(record)
            self.timer.start()
            return

        first_record = self.backlog[0]

        if (
            first_record.name == record.name
            and first_record.levelno == record.levelno
            and first_record.thread == record.thread
            and not first_record.args
            and not record.is_log_record
        ):
            self.backlog.append(record)
        else:
            return self.emitAll(record)

    def flush(self):
        if self.backlog:
            first_record = self.backlog[0]
            lines = [first_record.getMessage()]
            for record in self.backlog[1:]:
                lines.append(record.getMessage())
            first_record.msg = "\n".join(lines)
            self.emit_upstream(first_record)
            self.backlog = list()

    def close(self):
        self.timer.close()
        self.flush()


class RecognizingHandler(Handler):
    def __init__(
        self,
        handler: Handler,
        handler_name: str,
        recognizer: Optional[LogLineRecognizer] = None,
    ) -> None:
        super().__init__(handler.level)
        self.recognizer = recognizer or DefaultRecognizer()
        self.handler = handler
        self.handler_name: str = handler_name

    def recognize(self, record: ExtLogRecord) -> ExtLogRecord:
        log_line = self.recognizer.recognize(record.getMessage())
        record.is_log_record = log_line.level is not None

        if not record.is_log_record:
            return record

        assert log_line.level

        record.levelno = log_line.level.value
        record.levelname = log_line.level.name
        record.extra = {**record.extra, **log_line.attrs}
        return record

    def emit(self, log_record: LogRecord) -> None:
        record = ExtLogRecord.fromRecord(log_record)

        if record.args or record.suppress or record.is_from_encab:
            return self.handler.emit(record)

        self.handler.emit(self.recognize(record))


class ErrorHandler(Handler):
    def __init__(self, handler: Handler, handler_name: str) -> None:
        super().__init__(handler.level)
        self.handler = handler
        self.errors: int = 0
        self.handler_name: str = handler_name

    def emit(self, log_record: LogRecord) -> None:
        record = ExtLogRecord.fromRecord(log_record)

        if record.suppress:
            return

        try:
            self.handler.emit(record)
            if self.errors:
                self.errors = 0
        except (HTTPException, ConnectionError) as e:
            if not self.errors:
                mylogger.warning(
                    "GELF Handler %s failed to connect to Graylog:  %s",
                    self.handler_name,
                    str(e),
                    extra={"program": ENCAB_GELF, "suppress": True},
                )
            else:
                self.errors = (self.errors + 1) % 100
        except Exception as e:
            mylogger.exception(
                "GELF Handler %s:  %s",
                self.handler_name,
                str(e),
                extra={"program": ENCAB_GELF, "suppress": True},
            )
